@@ -1,23 +1,19 @@
-# -*- coding: utf-8 -*-
-
 """
 Tests for the path module.
 
-This suite runs on Linux, OS X, and Windows right now.  To extend the
+This suite runs on Linux, macOS, and Windows. To extend the
 platform support, just add appropriate pathnames for your
 platform (os.name) in each place where the p() function is called.
-Then report the result.  If you can't get the test to run at all on
-your platform, there's probably a bug in path.py -- please report the issue
-in the issue tracker at https://github.com/jaraco/path.py.
+Then report the result. If you can't get the test to run at all on
+your platform, there's probably a bug -- please report the issue
+in the issue tracker.
 
-TestScratchDir.test_touch() takes a while to run.  It sleeps a few
+TestScratchDir.test_touch() takes a while to run. It sleeps a few
 seconds to allow some time to pass between calls to check the modify
 time on files.
 """
 
-from __future__ import unicode_literals, absolute_import, print_function
-
-import codecs
+import io
 import os
 import sys
 import shutil
@@ -28,48 +24,30 @@ import posixpath
 import textwrap
 import platform
 import importlib
-import operator
 import datetime
 import subprocess
 import re
+import contextlib
+import stat
 
 import pytest
-import packaging.version
 
 import path
+from path import Path
 from path import TempDir
 from path import matchers
 from path import SpecialResolver
 from path import Multi
 
-Path = None
 
-
-def p(**choices):
-    """ Choose a value from several possible values, based on os.name """
+def os_choose(**choices):
+    """Choose a value from several possible values, based on os.name"""
     return choices[os.name]
 
 
-@pytest.fixture(autouse=True, params=[path.Path])
-def path_class(request, monkeypatch):
-    """
-    Invoke tests on any number of Path classes.
-    """
-    monkeypatch.setitem(globals(), 'Path', request.param)
-
-
-def mac_version(target, comparator=operator.ge):
-    """
-    Return True if on a Mac whose version passes the comparator.
-    """
-    current_ver = packaging.version.parse(platform.mac_ver()[0])
-    target_ver = packaging.version.parse(target)
-    return platform.system() == 'Darwin' and comparator(current_ver, target_ver)
-
-
 class TestBasics:
     def test_relpath(self):
-        root = Path(p(nt='C:\\', posix='/'))
+        root = Path(os_choose(nt='C:\\', posix='/'))
         foo = root / 'foo'
         quux = foo / 'quux'
         bar = foo / 'bar'
@@ -96,21 +74,15 @@ class TestBasics:
         cwd = Path(os.getcwd())
         assert boz.relpath() == cwd.relpathto(boz)
 
-        if os.name == 'nt':
+        if os.name == 'nt':  # pragma: nocover
             # Check relpath across drives.
             d = Path('D:\\')
             assert d.relpathto(boz) == boz
 
     def test_construction_from_none(self):
-        """
-
-        """
-        try:
+        """ """
+        with pytest.raises(TypeError):
             Path(None)
-        except TypeError:
-            pass
-        else:
-            raise Exception("DID NOT RAISE")
 
     def test_construction_from_int(self):
         """
@@ -119,7 +91,7 @@ class TestBasics:
         assert Path(1) == '1'
 
     def test_string_compatibility(self):
-        """ Test compatibility with ordinary strings. """
+        """Test compatibility with ordinary strings."""
         x = Path('xyzzy')
         assert x == 'xyzzy'
         assert x == str('xyzzy')
@@ -132,32 +104,33 @@ class TestBasics:
         # Test p1/p1.
         p1 = Path("foo")
         p2 = Path("bar")
-        assert p1 / p2 == p(nt='foo\\bar', posix='foo/bar')
+        assert p1 / p2 == os_choose(nt='foo\\bar', posix='foo/bar')
 
     def test_properties(self):
         # Create sample path object.
-        f = p(
-            nt='C:\\Program Files\\Python\\Lib\\xyzzy.py',
-            posix='/usr/local/python/lib/xyzzy.py',
+        f = Path(
+            os_choose(
+                nt='C:\\Program Files\\Python\\Lib\\xyzzy.py',
+                posix='/usr/local/python/lib/xyzzy.py',
+            )
         )
-        f = Path(f)
 
         # .parent
         nt_lib = 'C:\\Program Files\\Python\\Lib'
         posix_lib = '/usr/local/python/lib'
-        expected = p(nt=nt_lib, posix=posix_lib)
+        expected = os_choose(nt=nt_lib, posix=posix_lib)
         assert f.parent == expected
 
         # .name
         assert f.name == 'xyzzy.py'
-        assert f.parent.name == p(nt='Lib', posix='lib')
+        assert f.parent.name == os_choose(nt='Lib', posix='lib')
 
         # .ext
         assert f.ext == '.py'
         assert f.parent.ext == ''
 
         # .drive
-        assert f.drive == p(nt='C:', posix='')
+        assert f.drive == os_choose(nt='C:', posix='')
 
     def test_methods(self):
         # .abspath()
@@ -168,12 +141,6 @@ class TestBasics:
         assert isinstance(cwd, Path)
         assert cwd == os.getcwd()
 
-    def test_UNC(self):
-        if hasattr(os.path, 'splitunc'):
-            p = Path(r'\\python1\share1\dir1\file1.txt')
-            assert p.uncshare == r'\\python1\share1'
-            assert p.splitunc() == os.path.splitunc(str(p))
-
     def test_explicit_module(self):
         """
         The user may specify an explicit path module to use.
@@ -200,7 +167,7 @@ class TestBasics:
     def test_joinpath_on_instance(self):
         res = Path('foo')
         foo_bar = res.joinpath('bar')
-        assert foo_bar == p(nt='foo\\bar', posix='foo/bar')
+        assert foo_bar == os_choose(nt='foo\\bar', posix='foo/bar')
 
     def test_joinpath_to_nothing(self):
         res = Path('foo')
@@ -209,16 +176,12 @@ class TestBasics:
     def test_joinpath_on_class(self):
         "Construct a path from a series of strings"
         foo_bar = Path.joinpath('foo', 'bar')
-        assert foo_bar == p(nt='foo\\bar', posix='foo/bar')
+        assert foo_bar == os_choose(nt='foo\\bar', posix='foo/bar')
 
     def test_joinpath_fails_on_empty(self):
         "It doesn't make sense to join nothing at all"
-        try:
+        with pytest.raises(TypeError):
             Path.joinpath()
-        except TypeError:
-            pass
-        else:
-            raise Exception("did not raise")
 
     def test_joinpath_returns_same_type(self):
         path_posix = Path.using_module(posixpath)
@@ -228,6 +191,124 @@ class TestBasics:
         assert isinstance(res2, path_posix)
         assert res2 == 'foo/bar'
 
+    def test_radd_string(self):
+        res = 'foo' + Path('bar')
+        assert res == Path('foobar')
+
+    def test_fspath(self):
+        os.fspath(Path('foobar'))
+
+    def test_normpath(self):
+        assert Path('foo//bar').normpath() == os.path.normpath('foo//bar')
+
+    def test_expandvars(self, monkeypatch):
+        monkeypatch.setitem(os.environ, 'sub', 'value')
+        val = '$sub/$(sub)'
+        assert Path(val).expandvars() == os.path.expandvars(val)
+        assert 'value' in Path(val).expandvars()
+
+    def test_expand(self):
+        val = 'foobar'
+        expected = os.path.normpath(os.path.expanduser(os.path.expandvars(val)))
+        assert Path(val).expand() == expected
+
+    def test_splitdrive(self):
+        val = Path.using_module(ntpath)(r'C:\bar')
+        drive, rest = val.splitdrive()
+        assert drive == 'C:'
+        assert rest == r'\bar'
+        assert isinstance(rest, Path)
+
+    def test_relpathto(self):
+        source = Path.using_module(ntpath)(r'C:\foo')
+        dest = Path.using_module(ntpath)(r'D:\bar')
+        assert source.relpathto(dest) == dest
+
+    def test_walk_errors(self):
+        start = Path('/does-not-exist')
+        items = list(start.walk(errors='ignore'))
+        assert not items
+
+    def test_walk_child_error(self, tmpdir):
+        def simulate_access_denied(item):
+            if item.name == 'sub1':
+                raise OSError("Access denied")
+
+        p = Path(tmpdir)
+        (p / 'sub1').makedirs_p()
+        items = path.Traversal(simulate_access_denied)(p.walk(errors='ignore'))
+        assert list(items) == [p / 'sub1']
+
+    def test_read_md5(self, tmpdir):
+        target = Path(tmpdir) / 'some file'
+        target.write_text('quick brown fox and lazy dog')
+        assert target.read_md5() == b's\x15\rPOW\x7fYk\xa8\x8e\x00\x0b\xd7G\xf9'
+
+    def test_read_hexhash(self, tmpdir):
+        target = Path(tmpdir) / 'some file'
+        target.write_text('quick brown fox and lazy dog')
+        assert target.read_hexhash('md5') == '73150d504f577f596ba88e000bd747f9'
+
+    @pytest.mark.skipif("not hasattr(os, 'statvfs')")
+    def test_statvfs(self):
+        Path('.').statvfs()
+
+    @pytest.mark.skipif("not hasattr(os, 'pathconf')")
+    def test_pathconf(self):
+        assert isinstance(Path('.').pathconf(1), int)
+
+    def test_utime(self, tmpdir):
+        tmpfile = Path(tmpdir) / 'file'
+        tmpfile.touch()
+        new_time = (time.time() - 600,) * 2
+        assert Path(tmpfile).utime(new_time).stat().st_atime == new_time[0]
+
+    def test_chmod_str(self, tmpdir):
+        tmpfile = Path(tmpdir) / 'file'
+        tmpfile.touch()
+        tmpfile.chmod('o-r')
+        is_windows = platform.system() == 'Windows'
+        assert is_windows or not (tmpfile.stat().st_mode & stat.S_IROTH)
+
+    @pytest.mark.skipif("not hasattr(Path, 'chown')")
+    def test_chown(self, tmpdir):
+        tmpfile = Path(tmpdir) / 'file'
+        tmpfile.touch()
+        tmpfile.chown(os.getuid(), os.getgid())
+        import pwd
+
+        name = pwd.getpwuid(os.getuid()).pw_name
+        tmpfile.chown(name)
+
+    def test_renames(self, tmpdir):
+        tmpfile = Path(tmpdir) / 'file'
+        tmpfile.touch()
+        tmpfile.renames(Path(tmpdir) / 'foo' / 'alt')
+
+    def test_mkdir_p(self, tmpdir):
+        Path(tmpdir).mkdir_p()
+
+    def test_removedirs_p(self, tmpdir):
+        dir = Path(tmpdir) / 'somedir'
+        dir.mkdir()
+        (dir / 'file').touch()
+        (dir / 'sub').mkdir()
+        dir.removedirs_p()
+        assert dir.isdir()
+        assert (dir / 'file').isfile()
+        # TODO: shouldn't sub get removed?
+        # assert not (dir / 'sub').isdir()
+
+
+class TestReadWriteText:
+    @pytest.mark.filterwarnings('ignore:Writing bytes in write_text')
+    def test_read_write(self, tmpdir):
+        file = path.Path(tmpdir) / 'filename'
+        file.write_text('hello world')
+        assert file.read_text() == 'hello world'
+        assert file.read_bytes() == b'hello world'
+        file.write_text(b'hello world')
+
 
 class TestPerformance:
     @staticmethod
@@ -241,7 +322,7 @@ class TestPerformance:
 
     def test_import_time(self, monkeypatch):
         """
-        Import of path.py should take less than some limit.
+        Import should take less than some limit.
 
         Run tests in a subprocess to isolate from test suite overhead.
         """
@@ -252,6 +333,51 @@ class TestPerformance:
         assert duration < limit
 
 
+class TestOwnership:
+    def test_get_owner(self):
+        Path('/').get_owner()
+
+
+class TestLinks:
+    def test_link(self, tmpdir):
+        target = Path(tmpdir) / 'target'
+        target.write_text('hello', encoding='utf-8')
+        link = target.link(Path(tmpdir) / 'link')
+        assert link.read_text() == 'hello'
+
+    def test_symlink_none(self, tmpdir):
+        root = Path(tmpdir)
+        with root:
+            file = (Path('dir').mkdir() / 'file').touch()
+            file.symlink()
+            assert Path('file').isfile()
+
+    def test_readlinkabs_passthrough(self, tmpdir):
+        link = Path(tmpdir) / 'link'
+        Path('foo').abspath().symlink(link)
+        link.readlinkabs() == Path('foo').abspath()
+
+    def test_readlinkabs_rendered(self, tmpdir):
+        link = Path(tmpdir) / 'link'
+        Path('foo').symlink(link)
+        link.readlinkabs() == Path(tmpdir) / 'foo'
+
+
+class TestSymbolicLinksWalk:
+    def test_skip_symlinks(self, tmpdir):
+        root = Path(tmpdir)
+        sub = root / 'subdir'
+        sub.mkdir()
+        sub.symlink(root / 'link')
+        (sub / 'file').touch()
+        assert len(list(root.walk())) == 4
+
+        skip_links = path.Traversal(
+            lambda item: item.isdir() and not item.islink(),
+        )
+        assert len(list(skip_links(root.walk()))) == 3
+
+
 class TestSelfReturn:
     """
     Some methods don't necessarily return any value (e.g. makedirs,
@@ -290,6 +416,22 @@ class TestSelfReturn:
         assert p == ret
 
 
+@pytest.mark.skipif("not hasattr(Path, 'chroot')")
+def test_chroot(monkeypatch):
+    results = []
+    monkeypatch.setattr(os, 'chroot', results.append)
+    Path().chroot()
+    assert results == ['']
+
+
+@pytest.mark.skipif("not hasattr(Path, 'startfile')")
+def test_startfile(monkeypatch):
+    results = []
+    monkeypatch.setattr(os, 'startfile', results.append)
+    Path().startfile()
+    assert results == ['']
+
+
 class TestScratchDir:
     """
     Tests that run in a temporary directory (does not test TempDir class)
@@ -347,7 +489,7 @@ class TestScratchDir:
         assert t2 <= f.mtime <= t3
         if hasattr(os.path, 'getctime'):
             ct2 = f.ctime
-            if os.name == 'nt':
+            if platform.system() == 'Windows':  # pragma: nocover
                 # On Windows, "ctime" is CREATION time
                 assert ct == ct2
                 assert ct2 < t2
@@ -402,16 +544,14 @@ class TestScratchDir:
             assert files == files2
         finally:
             for f in files:
-                try:
+                with contextlib.suppress(Exception):
                     f.remove()
-                except Exception:
-                    pass
 
-    @pytest.mark.xfail(mac_version('10.13'), reason="macOS disallows invalid encodings")
-    @pytest.mark.xfail(
-        platform.system() == 'Windows', reason="Can't write latin characters. See #133"
+    @pytest.mark.skipif(
+        platform.system() != "Linux",
+        reason="Only Linux allows writing invalid encodings",
     )
-    def test_listdir_other_encoding(self, tmpdir):
+    def test_listdir_other_encoding(self, tmpdir):  # pragma: nocover
         """
         Some filesystems allow non-character sequences in path names.
         ``.listdir`` should still function in this case.
@@ -427,10 +567,10 @@ class TestScratchDir:
         # first demonstrate that os.listdir works
         assert os.listdir(tmpdir_bytes)
 
-        # now try with path.py
+        # now try with path
         results = Path(tmpdir).listdir()
         assert len(results) == 1
-        res, = results
+        (res,) = results
         assert isinstance(res, Path)
         # OS X seems to encode the bytes in the filename as %XX characters.
         if platform.system() == 'Darwin':
@@ -516,10 +656,7 @@ class TestScratchDir:
         assert testFile.bytes() == testCopy2.bytes()
 
         # Make a link for the next test to use.
-        if hasattr(os, 'symlink'):
-            testFile.symlink(testLink)
-        else:
-            testFile.copy(testLink)  # fallback
+        testFile.symlink(testLink)
 
         # Test copying directory tree.
         testA.copytree(testC)
@@ -543,7 +680,7 @@ class TestScratchDir:
         )
         if hasattr(os, 'symlink'):
             assert testCopyOfLink.islink()
-            assert testCopyOfLink.readlink() == testFile
+            assert testCopyOfLink.realpath() == testFile
 
         # Clean up.
         testDir.rmtree()
@@ -574,135 +711,112 @@ class TestScratchDir:
         self.assertList(d.walkfiles('*.tmp'), [e / 'x.tmp' for e in dirs])
         self.assertList(d.walkdirs('*.tmp'), [d / 'xdir.tmp'])
 
-    def test_unicode(self, tmpdir):
+    encodings = 'UTF-8', 'UTF-16BE', 'UTF-16LE', 'UTF-16'
+
+    @pytest.mark.parametrize("encoding", encodings)
+    def test_unicode(self, tmpdir, encoding):
+        """Test that path works with the specified encoding,
+        which must be capable of representing the entire range of
+        Unicode codepoints.
+        """
         d = Path(tmpdir)
         p = d / 'unicode.txt'
 
-        def test(enc):
-            """ Test that path works with the specified encoding,
-            which must be capable of representing the entire range of
-            Unicode codepoints.
-            """
+        givenLines = [
+            'Hello world\n',
+            '\u0d0a\u0a0d\u0d15\u0a15\r\n',
+            '\u0d0a\u0a0d\u0d15\u0a15\x85',
+            '\u0d0a\u0a0d\u0d15\u0a15\u2028',
+            '\r',
+            'hanging',
+        ]
+        given = ''.join(givenLines)
+        expectedLines = [
+            'Hello world\n',
+            '\u0d0a\u0a0d\u0d15\u0a15\n',
+            '\u0d0a\u0a0d\u0d15\u0a15\n',
+            '\u0d0a\u0a0d\u0d15\u0a15\n',
+            '\n',
+            'hanging',
+        ]
+        clean = ''.join(expectedLines)
+        stripped = [line.replace('\n', '') for line in expectedLines]
 
-            given = (
-                'Hello world\n'
-                '\u0d0a\u0a0d\u0d15\u0a15\r\n'
-                '\u0d0a\u0a0d\u0d15\u0a15\x85'
-                '\u0d0a\u0a0d\u0d15\u0a15\u2028'
-                '\r'
-                'hanging'
-            )
-            clean = (
-                'Hello world\n'
-                '\u0d0a\u0a0d\u0d15\u0a15\n'
-                '\u0d0a\u0a0d\u0d15\u0a15\n'
-                '\u0d0a\u0a0d\u0d15\u0a15\n'
-                '\n'
-                'hanging'
-            )
-            givenLines = [
-                ('Hello world\n'),
-                ('\u0d0a\u0a0d\u0d15\u0a15\r\n'),
-                ('\u0d0a\u0a0d\u0d15\u0a15\x85'),
-                ('\u0d0a\u0a0d\u0d15\u0a15\u2028'),
-                ('\r'),
-                ('hanging'),
-            ]
-            expectedLines = [
-                ('Hello world\n'),
-                ('\u0d0a\u0a0d\u0d15\u0a15\n'),
-                ('\u0d0a\u0a0d\u0d15\u0a15\n'),
-                ('\u0d0a\u0a0d\u0d15\u0a15\n'),
-                ('\n'),
-                ('hanging'),
-            ]
-            expectedLines2 = [
-                ('Hello world'),
-                ('\u0d0a\u0a0d\u0d15\u0a15'),
-                ('\u0d0a\u0a0d\u0d15\u0a15'),
-                ('\u0d0a\u0a0d\u0d15\u0a15'),
-                (''),
-                ('hanging'),
-            ]
+        # write bytes manually to file
+        with io.open(p, 'wb') as strm:
+            strm.write(given.encode(encoding))
 
-            # write bytes manually to file
-            f = codecs.open(p, 'w', enc)
-            f.write(given)
-            f.close()
+        # test all 3 path read-fully functions, including
+        # path.lines() in unicode mode.
+        assert p.bytes() == given.encode(encoding)
+        with pytest.deprecated_call():
+            assert p.text(encoding) == clean
+        assert p.lines(encoding) == expectedLines
+        assert p.lines(encoding, retain=False) == stripped
 
-            # test all 3 path read-fully functions, including
-            # path.lines() in unicode mode.
-            assert p.bytes() == given.encode(enc)
-            assert p.text(enc) == clean
-            assert p.lines(enc) == expectedLines
-            assert p.lines(enc, retain=False) == expectedLines2
+        # If this is UTF-16, that's enough.
+        # The rest of these will unfortunately fail because append=True
+        # mode causes an extra BOM to be written in the middle of the file.
+        # UTF-16 is the only encoding that has this problem.
+        if encoding == 'UTF-16':
+            return
 
-            # If this is UTF-16, that's enough.
-            # The rest of these will unfortunately fail because append=True
-            # mode causes an extra BOM to be written in the middle of the file.
-            # UTF-16 is the only encoding that has this problem.
-            if enc == 'UTF-16':
-                return
+        # Write Unicode to file using path.write_text().
+        # This test doesn't work with a hanging line.
+        cleanNoHanging = clean + '\n'
 
-            # Write Unicode to file using path.write_text().
-            # This test doesn't work with a hanging line.
-            cleanNoHanging = clean + '\n'
+        p.write_text(cleanNoHanging, encoding)
+        p.write_text(cleanNoHanging, encoding, append=True)
+        # Check the result.
+        expectedBytes = 2 * cleanNoHanging.replace('\n', os.linesep).encode(encoding)
+        expectedLinesNoHanging = expectedLines[:]
+        expectedLinesNoHanging[-1] += '\n'
+        assert p.bytes() == expectedBytes
+        with pytest.deprecated_call():
+            assert p.text(encoding) == 2 * cleanNoHanging
+        assert p.lines(encoding) == 2 * expectedLinesNoHanging
+        assert p.lines(encoding, retain=False) == 2 * stripped
 
-            p.write_text(cleanNoHanging, enc)
-            p.write_text(cleanNoHanging, enc, append=True)
-            # Check the result.
-            expectedBytes = 2 * cleanNoHanging.replace('\n', os.linesep).encode(enc)
-            expectedLinesNoHanging = expectedLines[:]
-            expectedLinesNoHanging[-1] += '\n'
-            assert p.bytes() == expectedBytes
-            assert p.text(enc) == 2 * cleanNoHanging
-            assert p.lines(enc) == 2 * expectedLinesNoHanging
-            assert p.lines(enc, retain=False) == 2 * expectedLines2
+        # Write Unicode to file using path.write_lines().
+        # The output in the file should be exactly the same as last time.
+        p.write_lines(expectedLines, encoding)
+        p.write_lines(stripped, encoding, append=True)
+        # Check the result.
+        assert p.bytes() == expectedBytes
 
-            # Write Unicode to file using path.write_lines().
-            # The output in the file should be exactly the same as last time.
-            p.write_lines(expectedLines, enc)
-            p.write_lines(expectedLines2, enc, append=True)
-            # Check the result.
-            assert p.bytes() == expectedBytes
+        # Now: same test, but using various newline sequences.
+        # If linesep is being properly applied, these will be converted
+        # to the platform standard newline sequence.
+        p.write_lines(givenLines, encoding)
+        p.write_lines(givenLines, encoding, append=True)
+        # Check the result.
+        assert p.bytes() == expectedBytes
 
-            # Now: same test, but using various newline sequences.
-            # If linesep is being properly applied, these will be converted
-            # to the platform standard newline sequence.
-            p.write_lines(givenLines, enc)
-            p.write_lines(givenLines, enc, append=True)
-            # Check the result.
-            assert p.bytes() == expectedBytes
+        # Same test, using newline sequences that are different
+        # from the platform default.
+        def testLinesep(eol):
+            p.write_lines(givenLines, encoding, linesep=eol)
+            p.write_lines(givenLines, encoding, linesep=eol, append=True)
+            expected = 2 * cleanNoHanging.replace('\n', eol).encode(encoding)
+            assert p.bytes() == expected
 
-            # Same test, using newline sequences that are different
-            # from the platform default.
-            def testLinesep(eol):
-                p.write_lines(givenLines, enc, linesep=eol)
-                p.write_lines(givenLines, enc, linesep=eol, append=True)
-                expected = 2 * cleanNoHanging.replace('\n', eol).encode(enc)
-                assert p.bytes() == expected
+        testLinesep('\n')
+        testLinesep('\r')
+        testLinesep('\r\n')
+        testLinesep('\x0d\x85')
 
-            testLinesep('\n')
-            testLinesep('\r')
-            testLinesep('\r\n')
-            testLinesep('\x0d\x85')
-
-            # Again, but with linesep=None.
-            p.write_lines(givenLines, enc, linesep=None)
-            p.write_lines(givenLines, enc, linesep=None, append=True)
-            # Check the result.
-            expectedBytes = 2 * given.encode(enc)
-            assert p.bytes() == expectedBytes
-            assert p.text(enc) == 2 * clean
-            expectedResultLines = expectedLines[:]
-            expectedResultLines[-1] += expectedLines[0]
-            expectedResultLines += expectedLines[1:]
-            assert p.lines(enc) == expectedResultLines
-
-        test('UTF-8')
-        test('UTF-16BE')
-        test('UTF-16LE')
-        test('UTF-16')
+        # Again, but with linesep=None.
+        p.write_lines(givenLines, encoding, linesep=None)
+        p.write_lines(givenLines, encoding, linesep=None, append=True)
+        # Check the result.
+        expectedBytes = 2 * given.encode(encoding)
+        assert p.bytes() == expectedBytes
+        with pytest.deprecated_call():
+            assert p.text(encoding) == 2 * clean
+        expectedResultLines = expectedLines[:]
+        expectedResultLines[-1] += expectedLines[0]
+        expectedResultLines += expectedLines[1:]
+        assert p.lines(encoding) == expectedResultLines
 
     def test_chunks(self, tmpdir):
         p = (TempDir() / 'test.txt').touch()
@@ -714,7 +828,6 @@ class TestScratchDir:
 
         assert i == len(txt) / size - 1
 
-    @pytest.mark.skipif(not hasattr(os.path, 'samefile'), reason="samefile not present")
     def test_samefile(self, tmpdir):
         f1 = (TempDir() / '1.txt').touch()
         f1.write_text('foo')
@@ -737,13 +850,12 @@ class TestScratchDir:
         (sub / 'afile').write_text('something')
         sub.rmtree_p()
         assert not sub.exists()
-        try:
-            sub.rmtree_p()
-        except OSError:
-            self.fail(
-                "Calling `rmtree_p` on non-existent directory "
-                "should not raise an exception."
-            )
+
+    def test_rmtree_p_nonexistent(self, tmpdir):
+        d = Path(tmpdir)
+        sub = d / 'subfolder'
+        assert not sub.exists()
+        sub.rmtree_p()
 
     def test_rmdir_p_exists(self, tmpdir):
         """
@@ -765,6 +877,18 @@ class TestScratchDir:
         assert not sub.exists()
         sub.rmdir_p()
 
+    def test_rmdir_p_sub_sub_dir(self, tmpdir):
+        """
+        A non-empty folder should not raise an exception.
+        """
+        d = Path(tmpdir)
+        sub = d / 'subfolder'
+        sub.mkdir()
+        subsub = sub / 'subfolder'
+        subsub.mkdir()
+
+        sub.rmdir_p()
+
 
 class TestMergeTree:
     @pytest.fixture(autouse=True)
@@ -781,10 +905,7 @@ class TestMergeTree:
         with open(self.test_file, 'w') as f:
             f.write('x' * 10000)
 
-        if hasattr(os, 'symlink'):
-            self.test_file.symlink(self.test_link)
-        else:
-            self.test_file.copy(self.test_link)
+        self.test_file.symlink(self.test_link)
 
     def check_link(self):
         target = Path(self.subdir_b / self.test_link.name)
@@ -854,12 +975,12 @@ class TestMergeTree:
         self.subdir_a.merge_tree(
             self.subdir_b, copy_function=path.only_newer(shutil.copy2)
         )
-        assert target.text() == 'this is newer'
+        assert target.read_text() == 'this is newer'
 
 
 class TestChdir:
     def test_chdir_or_cd(self, tmpdir):
-        """ tests the chdir or cd method """
+        """tests the chdir or cd method"""
         d = Path(str(tmpdir))
         cwd = d.getcwd()
 
@@ -933,27 +1054,23 @@ class TestTempDir:
         d.__exit__(None, None, None)
         assert not d.exists()
 
-    def test_context_manager_exception(self):
-        """
-        The context manager will not clean up if an exception occurs.
-        """
-        d = TempDir()
-        d.__enter__()
-        (d / 'somefile.txt').touch()
-        assert not isinstance(d / 'somefile.txt', TempDir)
-        d.__exit__(TypeError, TypeError('foo'), None)
-        assert d.exists()
-
     def test_context_manager_using_with(self):
         """
         The context manager will allow using the with keyword and
-        provide a temporry directory that will be deleted after that.
+        provide a temporary directory that will be deleted after that.
         """
 
         with TempDir() as d:
             assert d.isdir()
         assert not d.isdir()
 
+    def test_cleaned_up_on_interrupt(self):
+        with contextlib.suppress(KeyboardInterrupt):
+            with TempDir() as d:
+                raise KeyboardInterrupt()
+
+        assert not d.exists()
+
 
 class TestUnicode:
     @pytest.fixture(autouse=True)
@@ -1043,9 +1160,6 @@ class TestPatternMatching:
         assert p / 'sub1' / 'foo' / 'bar.Txt' in files
 
 
-@pytest.mark.skipif(
-    sys.version_info < (2, 6), reason="in_place requires io module in Python 2.6"
-)
 class TestInPlace:
     reference_content = textwrap.dedent(
         """
@@ -1100,6 +1214,11 @@ class TestInPlace:
         assert 'Lorem' not in data
         assert 'lazy dog' in data
 
+    def test_write_mode_invalid(self, tmpdir):
+        with pytest.raises(ValueError):
+            with (Path(tmpdir) / 'document').in_place(mode='w'):
+                pass
+
 
 class TestSpecialPaths:
     @pytest.fixture(autouse=True, scope='class')
@@ -1232,16 +1351,35 @@ class TestMultiPath:
 
 def test_no_dependencies():
     """
-    Path.py guarantees that the path module can be
+    Path pie guarantees that the path module can be
     transplanted into an environment without any dependencies.
     """
     cmd = [sys.executable, '-S', '-c', 'import path']
     subprocess.check_call(cmd)
 
 
-def test_version():
-    """
-    Under normal circumstances, path should present a
-    __version__.
-    """
-    assert re.match(r'\d+\.\d+.*', path.__version__)
+class TestHandlers:
+    @staticmethod
+    def run_with_handler(handler):
+        try:
+            raise ValueError()
+        except Exception:
+            handler("Something unexpected happened")
+
+    def test_raise(self):
+        handler = path.Handlers._resolve('strict')
+        with pytest.raises(ValueError):
+            self.run_with_handler(handler)
+
+    def test_warn(self):
+        handler = path.Handlers._resolve('warn')
+        with pytest.warns(path.TreeWalkWarning):
+            self.run_with_handler(handler)
+
+    def test_ignore(self):
+        handler = path.Handlers._resolve('ignore')
+        self.run_with_handler(handler)
+
+    def test_invalid_handler(self):
+        with pytest.raises(ValueError):
+            path.Handlers._resolve('raise')
